﻿//
// AssemblyUtils.cs
//
// Copyright (c) 2008 RednaxelaFX (Kris Mok, rednaxela0607@hotmail.com)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Reflection;

namespace ClassHierarchyViewer {
    public static class AssemblyUtils {

        /// <summary>
        /// Checks whether a path refers to a vliad .NET assembly.
        /// </summary>
        /// <param name="path">The path for the assembly to be checked.</param>
        /// <exception cref="System.ArgumentException">
        /// <i>path</i> is invalid, such as an assembly with an invalid culture.
        /// </exception>
        /// <exception cref="System.IO.FileNotFoundException">
        /// <i>path</i> is not found.
        /// </exception>
        /// <exception cref="System.Security.SecurityException">
        /// The caller does not have path discovery permission.
        /// </exception>
        /// <returns><b>true</b> if <i>path</i> refers to a valid assembly;
        /// otherwise <b>false</b>.</returns>
        public static bool IsAssemblyValid( string path ) {
            AssemblyName name = null;
            return TryGetAssemblyName( path, out name );
        }

        /// <summary>
        /// Gets the <b>AssemblyName</b> for a given file.
        /// </summary>
        /// <param name="path">The path for the assembly to be checked.</param>
        /// <param name="name">
        /// An <b>AssemblyName</b> object representing the given file.
        /// </param>
        /// <exception cref="System.ArgumentException">
        /// <i>path</i> is invalid, such as an assembly with an invalid culture.
        /// </exception>
        /// <exception cref="System.IO.FileNotFoundException">
        /// <i>path</i> is not found.
        /// </exception>
        /// <exception cref="System.Security.SecurityException">
        /// The caller does not have path discovery permission.
        /// </exception>
        /// <returns><b>true</b> if <i>path</i> refers to a valid assembly;
        /// otherwise <b>false</b>.</returns>
        public static bool TryGetAssemblyName( string path, out AssemblyName name ) {
            name = null;
            if ( string.IsNullOrEmpty( path ) ) return false;

            var result = true;
            try {
                // this call doesn't load the assembly into memory,
                // so we don't have to worry about memory usage problems.
                name = AssemblyName.GetAssemblyName( path );
            } catch ( BadImageFormatException ) {
                result = false;
            } catch ( FileLoadException ) {
                // Do nothing. If this exception is thrown, then the specified
                // file is a valid .NET assembly, just that it cannot be loaded,
                // due to: "An assembly or module was loaded twice with two
                // different sets of evidence."
            }

            return result;
        }
    }
}
