﻿//
// Helpers.cs
//
// Copyright (c) 2008 RednaxelaFX (Kris Mok, rednaxela0607@hotmail.com)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Windows.Forms;

namespace ClassHierarchyViewer {
    internal static class Helpers {
        private static readonly Type _objectType = typeof( object );
        private static readonly Type _enumType = typeof( Enum );
        private static readonly Type _valueType = typeof( ValueType );

        internal static void PopulateImageList( ImageList list ) {
            ImageKeyUtils.PopulateImageList( list );
        }

        private static void AddNestedTypeToClassList( TreeNode node, Type enclosingType ) {
            var bindingAttr =
                BindingFlags.Public | BindingFlags.NonPublic |
                BindingFlags.Instance | BindingFlags.Static;
            foreach ( var nestedType in enclosingType.GetNestedTypes( bindingAttr ) ) {
                var childNode = GetClassListNodeFromType( nestedType );
                AddNestedTypeToClassList( childNode, nestedType );
                node.Nodes.Add( childNode );
            }
        }

        internal static void PopulateClassList( TreeView view, Assembly assembly ) {
            view.BeginUpdate( );
            view.Nodes.Clear( );

            try {
                var types = assembly.GetTypes( );
                var namespaces = from t in types
                                 where !t.IsNested
                                 group t by t.Namespace into g
                                 orderby g.Key
                                 select g;
                foreach ( var ns in namespaces ) {
                    var nsNode = new TreeNode( ns.Key );
                    nsNode.ToolTipText = ns.Key;
                    nsNode.ImageKey = ImageKeys.Namespace;
                    nsNode.SelectedImageKey = ImageKeys.Namespace;
                    foreach ( var type in ns.OrderBy( t => t.Name ) ) {
                        var typeNode = GetClassListNodeFromType( type );
                        AddNestedTypeToClassList( typeNode, type );
                        nsNode.Nodes.Add( typeNode );
                    }
                    view.Nodes.Add( nsNode );
                }
            } catch ( Exception ex ) {
                MessageBox.Show( ex.ToString( ) );
            }

            view.EndUpdate( );
        }

        internal static void PopulateClassHierarchy( TreeView view, Type type, bool getFullName ) {
            view.BeginUpdate( );
            view.Nodes.Clear( );
            view.Tag = type;

            var typeNode = GetClassHierarchyNodeFromType( type, getFullName );
            TreeNode baseNode = null;
            TreeNode rootNode = null;
            if ( type.IsInterface ) {
                rootNode = typeNode;
                AddDerivedTypeNodes( typeNode, getFullName );
            } else if ( type.IsClass ) {
                if ( _objectType == type ) {
                    rootNode = typeNode;
                } else {
                    var baseType = type.BaseType;
                    baseNode = GetClassHierarchyNodeFromType( baseType, getFullName );
                    rootNode = baseNode;
                    while ( baseType != _objectType ) {
                        baseType = baseType.BaseType;
                        var newNode = GetClassHierarchyNodeFromType( baseType, getFullName );
                        newNode.Nodes.Add( rootNode );
                        rootNode = newNode;
                    }
                    baseNode.Nodes.Add( typeNode );
                }
                AddDerivedTypeNodes( typeNode, getFullName );
            } else if ( type.IsEnum ) {
                rootNode = GetClassHierarchyNodeFromType( _objectType, getFullName );
                baseNode = GetClassHierarchyNodeFromType( _enumType, getFullName );
                var valNode = GetClassHierarchyNodeFromType( _valueType, getFullName );
                rootNode.Nodes.Add( valNode );
                valNode.Nodes.Add( baseNode );
                baseNode.Nodes.Add( typeNode );
            } else { // type.IsValueType == true && type.IsEnum == false
                rootNode = GetClassHierarchyNodeFromType( _objectType, getFullName );
                baseNode = GetClassHierarchyNodeFromType( _valueType, getFullName );
                rootNode.Nodes.Add( baseNode );
                baseNode.Nodes.Add( typeNode );
            }

            view.Nodes.Add( rootNode );
            view.SelectedNode = typeNode;

            view.EndUpdate( );
        }

        internal static void PopulateMemberList( TreeView view, Type type, bool getFullName, bool useAlias ) {
            view.BeginUpdate( );
            view.Nodes.Clear( );
            view.Tag = type;

            foreach ( var member in GetMembers( type ) ) {
                var node = GetMemberNodeFromMember( member, getFullName, useAlias );
                view.Nodes.Add( node );
            }

            view.EndUpdate( );
        }

        internal static void AddDerivedTypeNodes( TreeNode rootNode, bool getFullName ) {
            foreach ( var type in GetDeriviedTypes( rootNode.Tag as Type ) ) {
                var childNode = GetClassHierarchyNodeFromType( type, getFullName );
                rootNode.Nodes.Add( childNode );
            }
        }

        static TreeNode GetClassListNodeFromType( Type type ) {
            var node = new TreeNode( );
            node.Text = GetTypeName( type, false, false );
            node.ToolTipText = GetTypeName( type, true, false );
            var imageKey = GetTypeImageKey( type );
            node.ImageKey = imageKey;
            node.SelectedImageKey = imageKey;
            node.Tag = type;
            if ( !type.IsVisible ) {
                node.ForeColor = Color.Gray;
            }
            return node;
        }

        static TreeNode GetClassHierarchyNodeFromType( Type type, bool getFullName ) {
            var node = new TreeNode( );
            node.Text = GetTypeName( type, getFullName, false );
            node.ToolTipText = GetTypeName( type, true, false );
            var imageKey = GetTypeImageKey( type );
            node.ImageKey = imageKey;
            node.SelectedImageKey = imageKey;
            node.Tag = type;
            return node;
        }

        static string GetTypeImageKey( Type type ) {
            return ImageKeyUtils.GetTypeImageKey( type );
        }

        static string GetTypeName( Type type, bool getFullName, bool useAlias ) {
            return type.GetTypeDescription( getFullName, useAlias );
        }

        static IEnumerable<Type> GetDeriviedTypes( Type type ) {
            return type.GetDerivedTypes( new[ ] { type.Assembly } );
        }

        static TreeNode GetMemberNodeFromMember( MemberInfo member, bool getFullName, bool useAlias ) {
            var node = new TreeNode( );
            node.Text = GetMemberName( member, getFullName, useAlias );
            node.ToolTipText = GetMemberName( member, true, false );
            var imageKey = GetMemberImageKey( member );
            node.ImageKey = imageKey;
            node.SelectedImageKey = imageKey;
            node.Tag = member;
            return node;
        }

        static string GetMemberImageKey( MemberInfo member ) {
            return ImageKeyUtils.GetMemberImageKey( member );
        }

        static IEnumerable<MemberInfo> GetMembers( Type type ) {
            return type.GetMembersOrdered( );
        }

        static string GetMemberName( MemberInfo member, bool getFullName, bool useAlias ) {
            var memberType = member.MemberType;
            switch ( memberType ) {
            case MemberTypes.Constructor:
                var c = member as ConstructorInfo;
                return GetConstructorName( c, getFullName, useAlias );
            case MemberTypes.Method:
                var m = member as MethodInfo;
                return string.Format(
                    "{0} : {1}",
                    GetMethodName( m, getFullName, useAlias ),
                    GetTypeName( m.ReturnType, getFullName, useAlias ) );
            case MemberTypes.Property:
                var p = member as PropertyInfo;
                return string.Format(
                    "{0} : {1}", p.Name, GetTypeName( p.PropertyType, getFullName, useAlias ) );
            case MemberTypes.Field:
                var f = member as FieldInfo;
                return string.Format(
                    "{0} : {1}", f.Name, GetTypeName( f.FieldType, getFullName, useAlias ) );
            case MemberTypes.Event:
                var e = member as EventInfo;
                return string.Format(
                    "{0} : {1}", e.Name, GetTypeName( e.EventHandlerType, getFullName, useAlias ) );
            }
            throw new Exception( "Should not reach here!" );
        }

        static string GetConstructorName( ConstructorInfo ctor, bool getFullName, bool useAlias ) {
            return ctor.GetConstructorDescription( getFullName, useAlias );
        }

        static string GetMethodName( MethodInfo method, bool getFullName, bool useAlias ) {
            return method.GetMethodDescription( getFullName, useAlias );
        }

        internal static string GetTextRepresentation(
            TreeNode node,
            Type rootType,
            bool getFullName,
            bool useAlias ) {

            var sb = new StringBuilder( );
            var view = node.TreeView;
            //view.
            view.BeginUpdate( );
            GetTextRepresentation( node, rootType, sb, 0, getFullName, useAlias );
            view.EndUpdate( );
            return sb.ToString( );
        }

        static void GetTextRepresentation(
            TreeNode node,
            Type rootType,
            StringBuilder sb,
            int indentLevel,
            bool getFullName,
            bool useAlias ) {

            var type = node.Tag as Type;
            if ( null == type ) return;

            bool isExpanded = node.IsExpanded;
            if ( type == rootType || type.IsSubclassOf( rootType ) || !isExpanded ) {
                node.Expand( );
            }

            var name = GetTypeName( type, getFullName, useAlias );
            sb.Append( new string( ' ', indentLevel * 2 ) );
            sb.AppendLine( name );

            foreach ( TreeNode child in node.Nodes ) {
                GetTextRepresentation( child, rootType, sb, indentLevel + 1, getFullName, useAlias );
            }
            if ( !isExpanded ) {
                node.Collapse( );
            }
        }
    }
}
