﻿//
// MainForm.cs
//
// Copyright (c) 2008 RednaxelaFX (Kris Mok, rednaxela0607@hotmail.com)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.ComponentModel;
using System.Reflection;
using System.Windows.Forms;

namespace ClassHierarchyViewer {
    public partial class MainForm : Form {

        private static readonly string _errMsgFile =
            "Failed to load the selected file as a .NET assembly";
        private static readonly string _errMsgName =
            "Failed to load the specified name as a .NET assembly";
        private static readonly string _errMsgTitle = "Error";

        #region Constructor

        public MainForm( string arg ) {
            InitializeComponent( );
            InitializeComponentExtra( );
            var path = ( arg ?? string.Empty ).Trim( );
            if ( string.Empty != path &&
               ( path.EndsWith( ".dll", StringComparison.CurrentCultureIgnoreCase )
              || path.EndsWith( ".exe", StringComparison.CurrentCultureIgnoreCase ) ) ) {
                if ( AssemblyUtils.IsAssemblyValid( path ) ) {
                    try {
                        LoadAssemblyFromPath( path );
                    } catch {
                        MessageBox.Show( _errMsgFile, _errMsgTitle );
                    }
                }
            }
        }

        #endregion

        #region Initialization & Loading

        private void InitializeComponentExtra( ) {
            Helpers.PopulateImageList( this.imageList );
        }

        private void LoadAssemblyFromString( ) {
            var name = this.comboBoxAssemblyName.Text;
            if ( null != name && string.Empty != name.Trim( ) ) {
                try {
                    // TODO: fix this, should use ReflectionOnly API instead
                    var assembly = Assembly.Load( name );
                    Helpers.PopulateClassList( this.treeViewClassList, assembly );
                    this.comboBoxAssemblyName.Text = assembly.FullName;
                } catch {
                    MessageBox.Show( _errMsgName, _errMsgTitle );
                }
            }
        }

        private void LoadAssemblyFromPath( string path ) {
            // TODO: fix this, should use ReflectionOnly API instead
            var assembly = Assembly.LoadFrom( path );
            Helpers.PopulateClassList( this.treeViewClassList, assembly );
            this.comboBoxAssemblyName.Text = assembly.FullName;
        }

        #endregion

        #region GUI Event Handlers

        private void buttonLoadAssemblyFromString_Click( object sender, EventArgs e ) {
            LoadAssemblyFromString( );
        }

        private void buttonLoadAssemblyFromFile_Click( object sender, EventArgs e ) {
            openFileDialogAssembly.ShowDialog( );
        }

        private void openFileDialogAssembly_FileOk( object sender, CancelEventArgs e ) {
            var path = this.openFileDialogAssembly.FileName;
            try {
                LoadAssemblyFromPath( path );
                e.Cancel = false;
            } catch {
                e.Cancel = true;
                MessageBox.Show( _errMsgFile, _errMsgTitle );
            }
        }

        private void checkBoxFullName_CheckedChanged( object sender, EventArgs e ) {
            var type = this.treeViewClassHierarchy.Tag as Type;
            if ( null != type ) {
                Helpers.PopulateClassHierarchy(
                    this.treeViewClassHierarchy,
                    type,
                    this.checkBoxFullName.Checked );
            }
        }

        private void checkBoxUseAlias_CheckedChanged( object sender, EventArgs e ) {
            var type = this.treeViewClassMember.Tag as Type;
            if ( null != type ) {
                Helpers.PopulateMemberList(
                    this.treeViewClassMember,
                    type,
                    this.checkBoxFullName.Checked,
                    this.checkBoxUseAlias.Checked );
            }
        }

        private void buttonExpandAll_Click( object sender, EventArgs e ) {
            this.treeViewClassHierarchy.BeginUpdate( );
            this.treeViewClassHierarchy.ExpandAll( );
            this.treeViewClassHierarchy.EndUpdate( );
        }

        private void buttonCopyToClipboard_Click( object sender, EventArgs e ) {
            var type = this.treeViewClassHierarchy.Tag as Type;
            if ( null == type ) return;

            var rootNode = this.treeViewClassHierarchy.Nodes[ 0 ];
            var getFullName = this.checkBoxFullName.Checked;
            var useAlias = this.checkBoxUseAlias.Checked;
            var text = Helpers.GetTextRepresentation(
                rootNode, type, getFullName, useAlias );
            Clipboard.SetText( text );
        }

        private void comboBoxAssemblyName_KeyDown( object sender, KeyEventArgs e ) {
            if ( e.KeyCode == Keys.Enter ) {
                e.SuppressKeyPress = true;
                LoadAssemblyFromString( );
            }
        }

        private void treeViewClassList_AfterSelect( object sender, TreeViewEventArgs e ) {
            var type = e.Node.Tag as Type;
            if ( null != type ) {
                Helpers.PopulateClassHierarchy(
                    this.treeViewClassHierarchy,
                    type,
                    this.checkBoxFullName.Checked );
            }
        }

        private void treeViewClassHierarchy_BeforeExpand( object sender, TreeViewCancelEventArgs e ) {
            var getFullName = this.checkBoxFullName.Checked;
            foreach ( TreeNode node in e.Node.Nodes ) {
                if ( 0 == node.Nodes.Count ) {
                    Helpers.AddDerivedTypeNodes( node, getFullName );
                }
            }
        }

        private void treeViewClassHierarchy_AfterSelect( object sender, TreeViewEventArgs e ) {
            var type = e.Node.Tag as Type;
            if ( null != type ) {
                Helpers.PopulateMemberList(
                    this.treeViewClassMember,
                    type,
                    this.checkBoxFullName.Checked,
                    this.checkBoxUseAlias.Checked );
            }
        }

        private void MainForm_DragEnter( object sender, DragEventArgs e ) {
            if ( e.Data.GetDataPresent( DataFormats.FileDrop ) ) {
                e.Effect = DragDropEffects.Copy;
            } else {
                e.Effect = DragDropEffects.None;
            }
        }

        private void MainForm_DragDrop( object sender, DragEventArgs e ) {
            var path = ( ( string[ ] ) e.Data.GetData( DataFormats.FileDrop ) )[ 0 ];
            if ( path.EndsWith( ".dll", StringComparison.CurrentCultureIgnoreCase )
              || path.EndsWith( ".exe", StringComparison.CurrentCultureIgnoreCase ) ) {
                try {
                    LoadAssemblyFromPath( path );
                } catch {
                    MessageBox.Show( _errMsgFile, _errMsgTitle );
                }
            }
        }

        #endregion
    }
}
