﻿//
// TypeUtils.cs
//
// Copyright (c) 2008 RednaxelaFX (Kris Mok, rednaxela0607@hotmail.com)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;

namespace ClassHierarchyViewer {
    internal enum Accessibility {
        Private,
        FamilyAndAssembly,
        Family,
        Assembly,
        FamilyOrAssembly,
        Public
    }

    internal static class TypeUtils {
        private static Dictionary<Type, string> _typeKeywords;

        private static readonly Type _objectType = typeof( object );
        private static readonly Type _enumType = typeof( Enum );
        private static readonly Type _valueType = typeof( ValueType );

        static TypeUtils( ) {
            _typeKeywords = new Dictionary<Type, string> {
                { typeof( object ), "object" },
                { typeof( string ), "string" },
                { typeof( char ),   "char"   },
                { typeof( bool ),   "bool"   },
                { typeof( byte ),   "byte"   },
                { typeof( sbyte ),  "sbyte"  },
                { typeof( short ),  "short"  },
                { typeof( ushort ), "ushort" },
                { typeof( int ),    "int"    },
                { typeof( uint ),   "uint"   },
                { typeof( long ),   "long"   },
                { typeof( ulong ),  "ulong"  },
                { typeof( float ),  "float"  },
                { typeof( double ), "double" },
                { typeof( void ),   "void"   }
            };
        }

        internal static Accessibility GetMethodAccessibility( MethodInfo method ) {
            if ( method.IsPublic ) {
                return Accessibility.Public;
            } else if ( method.IsAssembly ) {
                return Accessibility.Assembly;
            } else if ( method.IsFamily ) {
                return Accessibility.Family;
            } else if ( method.IsFamilyOrAssembly ) {
                return Accessibility.FamilyOrAssembly;
            } else if ( method.IsFamilyAndAssembly ) {
                return Accessibility.FamilyAndAssembly;
            } else {
                return Accessibility.Private;
            }
        }

        internal static IEnumerable<Type> GetDerivedTypes( this Type type, IEnumerable<Assembly> assemblies ) {
            if ( type.IsValueType ) {
                return Type.EmptyTypes;
            }
            if ( type.IsInterface ) {
                return from asm in assemblies
                       from t in asm.GetTypes( )
                       where t.GetInterfaces( ).Any(
                        i => type == ( i.IsGenericType ? i.GetGenericTypeDefinition( ) : i ) )
                       orderby t.IsInterface descending, t.Name
                       select t;
            } else { // type.IsClass == true
                return from asm in assemblies
                       from t in asm.GetTypes( )
                       where t.BaseType == type
                       orderby t.Name
                       select t;
            }
        }

        //static IEnumerable<Type> GetDerivedInterfaces( Type type ) {
        //    throw new NotImplementedException( );
        //}

        internal static string GetTypeDescription( this Type type, bool getFullName, bool useAlias ) {
            // return keyword alias if available
            if ( !getFullName && useAlias && _typeKeywords.ContainsKey( type ) ) {
                return _typeKeywords[ type ];
            }

            // get type's simple or full name
            string name;
            if ( getFullName && !type.IsGenericParameter ) {
                if ( type.IsNested ) {
                    name = string.Format( "{0}.{1}",
                        type.DeclaringType.GetTypeDescription( getFullName, useAlias ),
                        type.Name );
                } else {
                    name = type.FullName ?? type.ToString( );
                }
            } else if ( type.IsArray ) {
                name = string.Format( "{0}[{1}]",
                    type.GetElementType( ).GetTypeDescription( getFullName, useAlias ),
                    new string( ',', type.GetArrayRank( ) - 1 ) );
            } else {
                name = type.Name;
            }

            // deal with generic types
            var idx = -1;
            // TODO: review the logic for checking generics
            if ( type.IsGenericType && ( idx = name.IndexOf( '`' ) ) >= 0 ) {
                var sName = name.Substring( 0, idx );
                var genericTypeNames = new List<string>( );
                foreach ( var t in type.GetGenericArguments( ) ) {
                    genericTypeNames.Add( t.GetTypeDescription( getFullName, useAlias ) );
                }
                var genericName = string.Format( "{0}{1}{2}{3}",
                    sName,
                    "<",
                    string.Join( ", ", genericTypeNames.ToArray( ) ),
                    ">" );
                name = genericName;
            }

            return name;
        }

        internal static IEnumerable<MemberInfo> GetMembersOrdered( this Type type ) {
            var bindingAttr =
                BindingFlags.Public | BindingFlags.NonPublic |
                BindingFlags.Instance | BindingFlags.Static;
            return type.GetMembersOrdered( bindingAttr );
        }

        internal static IEnumerable<MemberInfo> GetMembersOrdered( this Type type, BindingFlags bindingAttr ) {
            var constructors = type.GetConstructors( bindingAttr )
                .OrderBy( c => c.GetParameters( ).Length ).Cast<MemberInfo>( );
            var methods = type.GetMethods( bindingAttr )
                .OrderBy( m => m.Name ).Cast<MemberInfo>( );
            var properties = type.GetProperties( bindingAttr )
                .OrderBy( p => p.Name ).Cast<MemberInfo>( );
            var fields = type.GetFields( bindingAttr )
                .OrderBy( f => f.Name ).Cast<MemberInfo>( );
            var events = type.GetEvents( bindingAttr )
                .OrderBy( e => e.Name ).Cast<MemberInfo>( );
            IEnumerable<MemberInfo> ret = new List<MemberInfo>( );
            if ( null != constructors ) ret = ret.Concat( constructors );
            if ( null != methods ) ret = ret.Concat( methods );
            if ( null != properties ) ret = ret.Concat( properties );
            if ( null != fields ) ret = ret.Concat( fields );
            if ( null != events ) ret = ret.Concat( events );
            return ret;
        }

        internal static string GetMemberDescription( MemberInfo member, bool getFullName ) {
            throw new NotImplementedException( );
        }

        internal static string GetMethodDescription( this MethodInfo method, bool getFullName ) {
            return method.GetMethodDescription( getFullName, true );
        }

        internal static string GetConstructorDescription( this ConstructorInfo ctor, bool getFullName, bool useAlias ) {
            var sb = new StringBuilder( );

            var name = ctor.Name;
            var idx = 0;
            // TODO: review the logic for checking generics
            if ( ctor.IsGenericMethod && ( idx = name.IndexOf( '`' ) ) >= 0 ) {
                var sName = name.Substring( 0, idx );
                var genericTypeNames = new List<string>( );
                foreach ( var t in ctor.GetGenericArguments( ) ) {
                    genericTypeNames.Add( t.GetTypeDescription( getFullName, useAlias ) );
                }
                sb.Append( sName );
                sb.AppendFormat( "{0}{1}{2}",
                    "<",
                    string.Join( ", ", genericTypeNames.ToArray<string>( ) ),
                    ">" );
            } else {
                sb.Append( name );
            }
            var paramNames = new List<string>( );
            foreach ( var param in ctor.GetParameters( ) ) {
                paramNames.Add( param.ParameterType.GetTypeDescription( getFullName, useAlias ) );
            }
            sb.AppendFormat( "({0})", string.Join( ", ", paramNames.ToArray<string>( ) ) );

            return sb.ToString( );
        }

        internal static string GetMethodDescription( this MethodInfo method, bool getFullName, bool useAlias ) {
            var sb = new StringBuilder( );

            var name = method.Name;
            var idx = -1;
            // TODO: review the logic for checking generics
            if ( method.IsGenericMethod && ( idx = name.IndexOf( '`' ) ) >= 0 ) {
                var sName = name.Substring( 0, idx );
                var genericTypeNames = new List<string>( );
                foreach ( var t in method.GetGenericArguments( ) ) {
                    genericTypeNames.Add( t.GetTypeDescription( getFullName, useAlias ) );
                }
                sb.AppendFormat( "{0}{1}{2}{3}",
                    sName,
                    "<",
                    string.Join( ", ", genericTypeNames.ToArray( ) ),
                    ">" );
            } else {
                sb.Append( name );
            }
            var paramNames = new List<string>( );
            foreach ( var param in method.GetParameters( ) ) {
                paramNames.Add( param.ParameterType.GetTypeDescription( getFullName, useAlias ) );
            }
            sb.AppendFormat( "({0})", string.Join( ", ", paramNames.ToArray<string>( ) ) );

            return sb.ToString( );
        }

        internal static string GetPropertyDescription( PropertyInfo property, bool getFullName ) {
            throw new NotImplementedException( );
        }

        internal static string GetFieldDescription( FieldInfo field, bool getFullName ) {
            throw new NotImplementedException( );
        }

        internal static string GetEventDescription( EventInfo ev, bool getFullName ) {
            throw new NotImplementedException( );
        }

        internal static bool IsTypeVisible( Type type ) {
            return type.IsVisible;
        }

        internal static bool IsMemberVisible( MemberInfo member ) {
            throw new NotImplementedException( );
        }
    }
}
